# -*- coding: utf-8 -*-
"""
State Model Unittests
=====================
Modified: 2022-03
"""

import builtins
import json
import unittest
import logging
from unittest.mock import MagicMock, patch, mock_open
from tests.resources.errors import JSON_DECODE_ERROR
from tests.resources.models import SERIALIZED_MODEL

from myosin.models.state import StateModel
from myosin.exceptions.cache import CachePathError, NullCachePathError


class TestStateModel(unittest.TestCase):

    @patch.multiple(
        StateModel,
        __abstractmethods__=set(),
        serialize=lambda x, y: SERIALIZED_MODEL,
        deserialize=lambda x, **y: None
    )
    def setUp(self) -> None:
        logging.disable()
        self.state = StateModel(1)  # type: ignore

    def tearDown(self) -> None:
        del self.state
        logging.disable(logging.NOTSET)

    def test_null_cache_path(self):
        """
        Test null cache path exception raise
        """
        with self.assertRaises(NullCachePathError):
            self.state.cache()

    def test_cache_path_existance(self):
        """
        Test cache path existance check
        """
        # inject a bad cache path
        self.state.cache_base_path = '\\'
        with self.assertRaises(CachePathError):
            self.state.cache()

    @patch.object(json, 'dump')
    @patch('builtins.open', mock_open())
    @patch.object(StateModel, 'serialize')
    def test_cache(self, dump: MagicMock, serialize: MagicMock):
        """
        Test caching mechanism
        """
        self.state.cache_base_path = '.'
        self.state.cache()
        dump.assert_called_once()
        serialize.assert_called_once()

    @patch.object(json, 'load')
    @patch('builtins.open', mock_open())
    def test_load_decode_error(self, load: MagicMock):
        """
        Test load decode error is handled gracefully
        """
        load.side_effect = JSON_DECODE_ERROR
        self.state.load()

    @patch.object(builtins, 'open')
    def test_load_fnf(self, open: MagicMock):
        """
        Test load file not found error is handled gracefully 
        """
        open.side_effect = FileNotFoundError
        self.state.load()

    @patch.object(json, 'load')
    @patch('builtins.open', mock_open())
    @patch.object(StateModel, 'deserialize')
    def test_load(self, load: MagicMock, deserialize: MagicMock):
        """
        Test loading mechanism
        """
        self.state.load()
        load.assert_called_once()
        deserialize.assert_called_once()



from myosin import StateModel
from typing import Any, Dict


SERIALIZED_MODEL = {
    'id': 1,
    'name': "cS",
    'email': "chris@email.com"
}


class DemoState(StateModel):

    def __init__(self, _id) -> None:
        super().__init__(_id)

    @property
    def name(self) -> str:
        return self.__name

    @name.setter
    def name(self, name: str) -> None:
        self.__name = name

    def serialize(self) -> Dict[str, Any]:
        return {
            'id': self.id,
            'name': self.name
        }

    def deserialize(self, **kwargs) -> None:
        for k, v in kwargs.items():
            setattr(self, k, v)



# -*- coding: utf-8 -*-

import logging
import asyncio
import traceback
from typing import Generic, List, Tuple, TypeVar, Callable

from myosin.models.state import StateModel
from myosin.typing import AsyncCallback


_S = TypeVar('_S', bound=StateModel)


class SSM(Generic[_S]):

    def __init__(self, reference: _S) -> None:
        self._logger = logging.getLogger(__name__)
        self.ref = reference
        self.queue = []

    @property
    def refhash(self) -> int:
        return hash(self.__ref)

    @property
    def typehash(self) -> int:
        return self.ref.__typehash__()

    @property
    def ref(self) -> _S:
        return self.__ref

    @ref.setter
    def ref(self, model: _S) -> None:
        self.__ref = model

    @property
    def queue(self) -> List[Callable[[_S], AsyncCallback]]:
        return self.__queue

    @queue.setter
    def queue(self, queue: List[Callable[[_S], AsyncCallback]]) -> None:
        self.__queue = queue

    async def execute(self) -> None:
        """
        Executes all subscriber coroutines with new model reference.
        """
        # construct coroutine lists
        tasks = [callback(self.ref) for callback in self.queue]
        # return results from coroutines with exceptions if any
        results = await asyncio.gather(*tasks, return_exceptions=True)
        # reformat tasks to display only the name
        operations = tuple(zip(map(lambda x: x.__qualname__, tasks), results))
        self._logger.debug("%s Subscriber operations: %s", type(self.ref).__name__, operations)
        # filter by operations which yielded an exception
        exceptions: List[Tuple[str, Exception]] = list(
            filter(lambda x: type(x[1]) is BaseException, operations))
        for func, exc in exceptions:
            self._logger.exception("Subscriber function: %s encountered an exception: %s", func,
                                   "".join(traceback.format_exception(
                                       etype=type(exc), value=exc, tb=exc.__traceback__
                                   )))



# -*- coding: utf-8 -*-
"""
State Exceptions
=====================
Modified: 2021-11
"""


class CacheException(Exception):
    def __init__(self, msg: str = "Cache encountered a general exception") -> None:
        self.message = msg


class CachePathError(CacheException):
    pass


class NullCachePathError(CacheException):
    pass



# -*- coding: utf-8 -*-
"""
State Exceptions
=====================
Modified: 2021-11
"""


class StateException(Exception):
    def __init__(self, msg: str = "StateIO encountered a general exception") -> None:
        self.message = msg


class UninitializedStateError(StateException):
    def __init__(self, msg: str = "") -> None:
        super().__init__(msg)


class HashNotFound(StateException):
    def __init__(self, msg: str = "The input model has an unexpected typehash") -> None:
        super().__init__(msg=msg)


class NullCheckoutError(StateException):
    def __init__(self, msg: str = "The requested model does not exist in state registry") -> None:
        super().__init__(msg=msg)



import os
import uuid
import logging
from threading import Thread
import time
from typing import Any, Dict
from datetime import datetime

from myosin import _PKey
from myosin import StateModel
from myosin import State


class User(StateModel):

    def __init__(self, _id: _PKey) -> None:
        super().__init__(_id)

    @property
    def name(self) -> str:
        return self.__name

    @name.setter
    def name(self, name: str) -> None:
        self.__name = name

    @property
    def email(self) -> str:
        return self.__email

    @email.setter
    def email(self, email: str) -> None:
        self.__email = email

    def serialize(self) -> Dict[str, Any]:
        return {
            'id': self.id,
            'name': self.name,
            'email': self.email
        }

    def deserialize(self, **kwargs) -> None:
        for k, v in kwargs.items():
            setattr(self, k, v)


class Telemetry(StateModel):

    def __init__(self, _id: _PKey) -> None:
        super().__init__(_id)

    @property
    def tp(self) -> float:
        return self.__tp

    @tp.setter
    def tp(self, value: float) -> None:
        self.__tp = value

    @property
    def timestamp(self) -> float:
        return datetime.now().timestamp()

    def serialize(self) -> Dict[str, Any]:
        return {
            'id': self.id,
            'temp': self.tp,
            'timestamp': self.timestamp
        }

    def deserialize(self, **kwargs) -> None:
        for k, v in kwargs.items():
            if k == "timestamp":
                continue
            setattr(self, k, v)


class Runner:
    def __init__(self) -> None:
        self.username = ""
        with State() as state:
            state.subscribe(User, self.update)

    async def update(self, usr: User) -> None:
        self.username = usr.name

    def run(self):
        while True:
            time.sleep(0.001)
            print(f"This is my username: {self.username}")


if __name__ == "__main__":
    os.environ['MYOSIN_CACHE_BASE_PATH'] = "tests/cache"
    logging.basicConfig(level=logging.DEBUG)
    _log = logging.getLogger(__name__)

    u = User(_id=1)
    u.deserialize(**{'name': "chris", 'email': "chris@email.io"})
    t = Telemetry(_id=1)
    t.deserialize(**{'tp': 65.0})
    u.email = "bad"
    _log.info(u)

    with State() as state:
        state.load(u)
        state.load(t)

    with State() as state:
        user = state.checkout(User)
        _log.info(user.email)

    with State() as state:
        user = state.checkout(User)
        user.email = "bad@email.com"
        state.commit(user, cache=True)

    with State() as state:
        user = state.checkout(User)
        _log.info(user)

    with State() as state:
        telemetry = state.checkout(Telemetry)
        _log.info(telemetry)

    with State() as state:
        telemetry = state.checkout(Telemetry)
        telemetry.tp = 45.0
        state.commit(telemetry, cache=True)

    # test state model subscriptions
    _runner = Runner()
    Thread(target=_runner.run, args=(), daemon=True).start()
    while True:
        time.sleep(0.001)
        with State() as state:
            usr = state.checkout(User)
            usr.name = str(uuid.uuid4())
            state.commit(usr)



# -*- coding: utf-8 -*-

import io
import os
from setuptools import find_packages, setup

# meta-data.
NAME = 'myosin'
DESCRIPTION = 'Lightweight state management engine for posix compliant systems'
EMAIL = 'christian@leapsystems.online'
AUTHOR = 'Christian Sargusingh'
REQUIRES_PYTHON = '>=3.7.0'
VERSION = False

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'requirements.txt'), 'r') as requirements:
    REQUIRED = list()
    for line in requirements.readlines():
        REQUIRED.append(line)

# Import the README and use it as the long-description.
try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    project_slug = "myosin"
    with open(os.path.join(here, project_slug, '__version__.py')) as f:
        exec(f.read(), about)
else:
    about['__version__'] = VERSION

setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    url='https://github.com/ztnel/myosin',
    python_requires=REQUIRES_PYTHON,
    packages=find_packages(),
    include_package_data=True,
    setup_requires=['wheel'],
    install_requires=REQUIRED,
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Operating System :: Unix',
        'Topic :: Software Development :: Object Brokering',
        'Topic :: Software Development :: Embedded Systems',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    ],
)
